@@ -50,6 +50,7 @@ static void     gail_focus_notify        (GtkWidget            *widget);
 static void     gail_focus_notify_when_idle (GtkWidget            *widget);
 
 static void     gail_focus_tracker_init (void);
+static void     gail_focus_object_destroyed (gpointer data);
 static void     gail_focus_tracker (AtkObject *object);
 
 static GtkWidget* focus_widget = NULL;
@@ -204,8 +205,9 @@ gail_focus_watcher (GSignalInvocationHint *ihint,
                       if (GTK_IS_MENU_ITEM (next_focus_widget) &&
                           !focus_before_menu)
                         {
+                          void *vp_focus_before_menu = &focus_before_menu;
                           focus_before_menu = window->focus_widget;
-                          g_object_add_weak_pointer (G_OBJECT (focus_before_menu), (gpointer *)&focus_before_menu);
+                          g_object_add_weak_pointer (G_OBJECT (focus_before_menu), vp_focus_before_menu);
                         }
 
                       return TRUE;
@@ -345,8 +347,9 @@ gail_finish_select (GtkWidget *widget)
           (GTK_IS_MENU_BAR (next_focus_widget) ||
            GTK_IS_MENU_ITEM (next_focus_widget)))
         {
+          void *vp_next_focus_widget = &next_focus_widget;
           g_source_remove (focus_notify_handler);
-          g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), (gpointer *)&next_focus_widget);
+          g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
           focus_notify_handler = 0;
           was_deselect = FALSE;
         }
@@ -359,8 +362,9 @@ gail_finish_select (GtkWidget *widget)
       !GTK_IS_MENU_ITEM (focus_widget) && 
       !GTK_IS_MENU (focus_widget))
     {
+      void *vp_focus_before_menu = &focus_before_menu;
       focus_before_menu = focus_widget;
-      g_object_add_weak_pointer (G_OBJECT (focus_before_menu), (gpointer *)&focus_before_menu);
+      g_object_add_weak_pointer (G_OBJECT (focus_before_menu), vp_focus_before_menu);
 
     } 
   gail_focus_notify_when_idle (widget);
@@ -475,7 +479,8 @@ gail_focus_idle_handler (gpointer data)
     }
   else
     {
-      g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), (gpointer *)&next_focus_widget);
+      void *vp_next_focus_widget = &next_focus_widget;
+      g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
       next_focus_widget = NULL;
     }
     
@@ -494,19 +499,22 @@ gail_focus_notify (GtkWidget *widget)
     {
       if (focus_widget)
         {
-          g_object_remove_weak_pointer (G_OBJECT (focus_widget), (gpointer *)&focus_widget);
+          void *vp_focus_widget = &focus_widget;
+          g_object_remove_weak_pointer (G_OBJECT (focus_widget), vp_focus_widget);
         }
       focus_widget = widget;
       if (focus_widget)
         {
-          g_object_add_weak_pointer (G_OBJECT (focus_widget), (gpointer *)&focus_widget);
+          void *vp_focus_widget = &focus_widget;
+          g_object_add_weak_pointer (G_OBJECT (focus_widget), vp_focus_widget);
           /*
            * The UI may not have been updated yet; e.g. in gtkhtml2
            * html_view_layout() is called in a idle handler
            */
           if (focus_widget == focus_before_menu)
             {
-              g_object_remove_weak_pointer (G_OBJECT (focus_before_menu), (gpointer *)&focus_before_menu);
+              void *vp_focus_before_menu = &focus_before_menu;
+              g_object_remove_weak_pointer (G_OBJECT (focus_before_menu), vp_focus_before_menu);
               focus_before_menu = NULL;
             }
         }
@@ -559,7 +567,10 @@ gail_focus_notify_when_idle (GtkWidget *widget)
             }
           g_source_remove (focus_notify_handler);
           if (next_focus_widget)
-            g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), (gpointer *)&next_focus_widget);
+	    {
+	      void *vp_next_focus_widget = &next_focus_widget;
+	      g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
+	    }
         }
       else
         /*
@@ -570,8 +581,9 @@ gail_focus_notify_when_idle (GtkWidget *widget)
 
   if (widget)
     {
+      void *vp_next_focus_widget = &next_focus_widget;
       next_focus_widget = widget;
-      g_object_add_weak_pointer (G_OBJECT (next_focus_widget), (gpointer *)&next_focus_widget);
+      g_object_add_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
     }
   focus_notify_handler = g_idle_add (gail_focus_idle_handler, widget);
 }
@@ -606,8 +618,9 @@ gail_deactivate_watcher (GSignalInvocationHint *ihint,
       (GTK_IS_MENU_BAR (next_focus_widget) ||
        GTK_IS_MENU_ITEM (next_focus_widget)))
     {
+      void *vp_next_focus_widget = &next_focus_widget;
       g_source_remove (focus_notify_handler);
-      g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), (gpointer *)&next_focus_widget);
+      g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), vp_next_focus_widget);
       focus_notify_handler = 0;
       was_deselect = FALSE;
     }
@@ -674,6 +687,16 @@ gail_focus_tracker_init ()
     }
 }
 
+static void
+gail_focus_object_destroyed (gpointer data)
+{
+  GObject *obj;
+
+  obj = G_OBJECT (data);
+  g_object_set_qdata (obj, quark_focus_object, NULL);
+  g_object_unref (obj); 
+}
+
 static void
 gail_focus_tracker (AtkObject *focus_object)
 {
@@ -700,6 +723,22 @@ gail_focus_tracker (AtkObject *focus_object)
               old_focus_object = g_object_get_qdata (G_OBJECT (parent), quark_focus_object);
               if (old_focus_object != focus_object)
                 {
+                  if (old_focus_object)
+                    g_object_weak_unref (G_OBJECT (old_focus_object),
+                                         (GWeakNotify) gail_focus_object_destroyed,
+                                         parent);
+                  else
+                    /*
+                     * We call g_object_ref as if the parent is destroyed 
+                     * while the weak reference exists then destroying the 
+                     * child would cause gail_focus_object_destroyed to be 
+                     * called when parent is not a valid GObject.
+                     */
+                    g_object_ref (parent);
+
+                  g_object_weak_ref (G_OBJECT (focus_object),
+                                     (GWeakNotify) gail_focus_object_destroyed,
+                                     parent);
                   g_object_set_qdata (G_OBJECT (parent), quark_focus_object, focus_object);
                 }
             }
@@ -709,7 +748,11 @@ gail_focus_tracker (AtkObject *focus_object)
           old_focus_object = g_object_get_qdata (G_OBJECT (focus_object), quark_focus_object);
           if (old_focus_object)
             {
+              g_object_weak_unref (G_OBJECT (old_focus_object),
+                                   (GWeakNotify) gail_focus_object_destroyed,
+                                   focus_object);
               g_object_set_qdata (G_OBJECT (focus_object), quark_focus_object, NULL);
+              g_object_unref (G_OBJECT (focus_object));
             }
         }
     }
